#! /usr/bin/env python
###############################################################################
# so.py                                                                       #
# Roger Whittaker                                                             #
# 20050504                                                                    #
# SUDOKU problem solver                                                       #
###############################################################################
# The python files required are:                                              #
#   parts.py                                                                  #
#   readfile.py                                                               #
#   solve.py                                                                  #
#   so.py                                                                     #
###############################################################################
# Copyright Roger Whittaker 2005                                              #
# This file may be freely used, modified and redistributed provided that this #
# copyright notice is retained                                                #
###############################################################################
from pprint   import pprint
###############################################################################
from readfile import readfile, display, writefile
from parts    import A, B, C, R, I, N, RR
###############################################################################
# A is all cells  (i,j) for 0 <= i <= 8, 0 <= j <= 8                          #
# B is all blocks                                                             #
# C is all columns                                                            #
# R is all rows                                                               #
# I is all items: the sum of the lists, B, C and R                            #
# N is ['1', '2', '3', '4', '5', '6', '7', '8', '9']                          #
# RR(n, i) is the set of n-tuples of distinct elements from i                 #
###############################################################################
def rowof(tup):
    return set(R[tup[0]])
###############################################################################
def colof(tup):
    return set(C[tup[1]])
###############################################################################
def boxof(tup):
    for box in B:
        if tup in box:
            ans = set(box)
    return ans
###############################################################################
def allof(tup):
    return boxof(tup)|rowof(tup)|colof(tup) 
###############################################################################
def uniq(item, dict):
    d = {}
    for i in A:
        if dict[i] == set([item]):
            d[i] = item
    return d
###############################################################################
class so:
    """
This defines the sudoku object
"""
    def __init__(self, d):
        self.d = d
        #######################################################################
        # d is the dictionary of actual values per cell.                      #
        #######################################################################
        return None

    def keys(self):
        return self.d.keys()

    def known(self):
        k = 0
        for key in self.d.keys():
            if self.d[key] != '':
                k = k + 1
        return k

    def __getitem__(self, key):
        return self.d[key]

    def __setitem__(self, key, value):
        self.d[key] = value
        return self
    
    def __repr__(self):
        return display(self)

    def valid(self):
        check = 0
        for key in self.celldict.keys():
            if key not in A:
                check = check + 1
        if check == 0:
            return 1
        else:
            return 0

    def possible(self):
        d = self.d
        #######################################################################
        # n is a copy of d which we output once updated.                      #
        #######################################################################
        n = {}
        for i in A:
            n[i] = ''
        for i in d.keys():
            n[i] = d[i]
        #######################################################################
        # p is the dictionary of possible values per cell.                    #
        #######################################################################
        p = {}
        for i in A:
            if d[i] != ['']:
                p[i] = set(N)
            else:
                p[i] = set([])
        #######################################################################
        # Basic removal of knowns from the values of p.                       #
        #######################################################################
        for i in d.keys():
            for j in allof(i):
                if j!= i:
                    if self.d[i] in p[j]:
                        if len(p[j]) != 1:
                            p[j].remove(d[i])
        #######################################################################
        # If we find that a value is unique in the possibles for an item,     #
        # it must be the correct one.                                         #
        #######################################################################
        def remuniq():
            for item in N:
                u = uniq(item, p)
                for key in u.keys():
                    for j in allof(key):
                        if j != key:
                            if u[key] in p[j]:
                                p[j].remove(u[key])
        #######################################################################
        # The set if unknown cells in the current item with between 2 and n   #
        # possibles.                                                          #
        #######################################################################
        def nof(rowcolbox, n):
            empties = [x for x in rowcolbox if d[x] == '']
            nof = set([x for x in empties if len(p[x]) <= n and len(p[x]) > 1])
            return nof
        #######################################################################
        # Go through all sets of n cells, and if the union of the possibles   #
        # is of length n, we can remove the possibles from those not in the   #
        # set.                                                                #
        #######################################################################
        def purge(collection, n, function):
            for item in collection:
                s = RR(n, nof(item, n))
                if s:
                    for i in s:
                        U = set([])
                        for j in i:
                            U|=p[j]
                        if len(U) == n:
                            for u in U:
                                for k in function(i[0]):
                                    if k not in i:
                                        if u in p[k]:
                                            p[k].remove(u)
        #######################################################################
        # The set of cells in the item for which num is a possible.           #
        #######################################################################
        def occur(num, item):
            o = set([])
            for j in item:
                if num in p[j]:
                    o.add(j)
            return o
        #######################################################################
        def update():
            for a in A:
                if len(p[a]) == 0:
                    ###########################################################
                    # We should never have an empty p[a].                     #
                    # If we do it means the sudoku is impossible.             #
                    ###########################################################
                    print 'ERROR'
                if len(p[a]) == 1:
                    n[a] = list(p[a])[0]
        #######################################################################
        remuniq()
        #######################################################################
        for j in [2, 3, 4, 5, 6, 7]:
            for (T, t) in [(R, rowof), (C, colof), (B, boxof)]:
                purge(T, j, t)
        #######################################################################
        # If all the occurances of num as possibilities in X are also in Y,   #
        # then we can remove them as possibilities from the rest of Y.        #
        # This part is crucial in solving the harder puzzles.                 #
        #######################################################################
        for num in N:
            for (X, Y) in [(B, R), (B, C), (R, C), (R, B), (C ,R), (C, B)]:
                for x in X:
                    for y in Y:
                        if occur(num, x) <= set(y):
                            for j in set(y) - set(x):
                                if num in p[j]:
                                    p[j].remove(num)
        #######################################################################
        update()
        #######################################################################
        return p , n
###############################################################################
if __name__ == '__main__':
    print 'testing so.py'
    print 
    print '    to solve problems run solve.py'
    print
    print '    sample file: sudoku.100'
    a = so(readfile('sudoku.100'))
    print '    a is the sudoku:' 
    print 
    print `a`
    print '    a.possible() is'
    print
    pprint(a.possible())
###############################################################################
